@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2021-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -25,10 +25,12 @@
 #include "kernel/core/locks.h"
 #include "gpu/subdevice/subdevice.h"
 #include "kernel/gpu/mem_mgr/heap.h"
+#include "gpu/mem_mgr/phys_mem_allocator/phys_mem_allocator.h"
 #include "kernel/gpu/mem_mgr/mem_mgr.h"
 #include "kernel/gpu/mig_mgr/kernel_mig_manager.h"
 #include "kernel/gpu/rc/kernel_rc.h"
 #include "kernel/gpu/bif/kernel_bif.h"
+#include "kernel/gpu/bus/kern_bus.h"
 #include "kernel/os/os.h"
 #include "platform/sli/sli.h"
 
@@ -486,7 +488,7 @@ krcWatchdogInit_IMPL
             Heap *pHeap = GPU_GET_HEAP(pGpu);
             NvU32 pmaConfig = PMA_QUERY_NUMA_ENABLED | PMA_QUERY_NUMA_ONLINED;
 
-            if (pmaQueryConfigs(&pHeap->pmaObject, &pmaConfig) == NV_OK)
+            if (pmaQueryConfigs(pHeap->pPmaObject, &pmaConfig) == NV_OK)
             {
                 // PMA can't be used until it's onlined
                 if (pmaConfig & PMA_QUERY_NUMA_ENABLED)
@@ -548,6 +550,9 @@ krcWatchdogInit_IMPL
         bAcquireLock = NV_FALSE;
     }
 
+    // Due to the Bug 4088184 WAR above, watchdog has to be explicitly set to internal
+    rmclientSetClientFlagsByHandle(hClient, RMAPI_CLIENT_FLAG_RM_INTERNAL_CLIENT);
+
     // Alloc device
     {
         NV0080_ALLOC_PARAMETERS *pNv0080 = &pParams->nv0080;
@@ -698,6 +703,7 @@ krcWatchdogInit_IMPL
     {
         NV_MEMORY_ALLOCATION_PARAMS *pMem = &pParams->mem;
         NvU32 hClass = NV01_MEMORY_SYSTEM;
+        NVOS46_PARAMETERS mapDmaParams = {0};
 
         portMemSet(pMem, 0, sizeof *pMem);
         pMem->owner = HEAP_OWNER_RM_CLIENT_GENERIC;
@@ -817,24 +823,24 @@ krcWatchdogInit_IMPL
 
         portMemSet(pKernelRc->watchdogChannelInfo.pCpuAddr, 0, pMem->size);
 
+        mapDmaParams.hClient   = hClient;
+        mapDmaParams.hDevice   = WATCHDOG_DEVICE_ID;
+        mapDmaParams.hDma      = WATCHDOG_VIRTUAL_CTX_ID;
+        mapDmaParams.hMemory   = WATCHDOG_MEM_ID;
+        mapDmaParams.length    = allocationSize;
+        mapDmaParams.flags     = (bCacheSnoop ? DRF_DEF(OS46, _FLAGS, _CACHE_SNOOP, _ENABLE) :
+                                     DRF_DEF(OS46, _FLAGS, _CACHE_SNOOP, _DISABLE)) |
+                                     DRF_DEF(OS46, _FLAGS, _ACCESS, _READ_WRITE);
         // Map the allocation into the unified heap.
-        status = pRmApi->Map(pRmApi,
-            hClient                 /* hClient */,
-            WATCHDOG_DEVICE_ID      /* hDevice */,
-            WATCHDOG_VIRTUAL_CTX_ID /* hMemctx */,
-            WATCHDOG_MEM_ID         /* hMemory */,
-            0                       /* offset */,
-            allocationSize          /* length */,
-            (bCacheSnoop ? DRF_DEF(OS46, _FLAGS, _CACHE_SNOOP, _ENABLE) :
-                           DRF_DEF(OS46, _FLAGS, _CACHE_SNOOP, _DISABLE)) |
-                DRF_DEF(OS46, _FLAGS, _ACCESS, _READ_WRITE),
-            &pKernelRc->watchdogChannelInfo.pGpuAddr);
+        status = pRmApi->Map(pRmApi, &mapDmaParams);
+
         if (status != NV_OK)
         {
             NV_PRINTF(LEVEL_ERROR,
                       "Unable to map memory into watchdog's heap\n");
             goto error;
         }
+        pKernelRc->watchdogChannelInfo.pGpuAddr = mapDmaParams.dmaOffset;
     }
 
     // Allocate the error notifier context DMA.
@@ -956,6 +962,7 @@ krcWatchdogInit_IMPL
         // Using device handle since VGPU doesnt support subdevice memory
         // allocations
         //
+        pMem->attr |= DRF_DEF(OS32, _ATTR, _ALLOCATE_FROM_RESERVED_HEAP, _YES);
         status = pRmApi->AllocWithHandle(pRmApi,
             hClient                    /* hClient */,
             WATCHDOG_DEVICE_ID         /* hParent */,
@@ -1385,9 +1392,17 @@ krcWatchdogInitPushbuffer_IMPL
 
     SLI_LOOP_START(SLI_LOOP_FLAGS_NONE);
     {
+        //
+        // On some architectures, if doorbell is mapped via bar0, we need to send
+        // an extra flush
+        //
+        if (kbusFlushPcieForBar0Doorbell_HAL(pGpu, GPU_GET_KERNEL_BUS(pGpu)) != NV_OK)
+        {
+            NV_PRINTF(LEVEL_ERROR, "Busflush failed.\n");
+            return;
+        }
         kfifoUpdateUsermodeDoorbell_HAL(pGpu, GPU_GET_KERNEL_FIFO(pGpu),
-            pKernelRc->watchdog.notifierToken->info32,
-            pKernelRc->watchdog.runlistId);
+            pKernelRc->watchdog.notifierToken->info32);
     }
     SLI_LOOP_END;
 
@@ -1457,8 +1472,7 @@ krcWatchdogWriteNotifierToGpfifo_IMPL
     SLI_LOOP_START(SLI_LOOP_FLAGS_NONE);
     {
         kfifoUpdateUsermodeDoorbell_HAL(pGpu, GPU_GET_KERNEL_FIFO(pGpu),
-            pKernelRc->watchdog.notifierToken->info32,
-            pKernelRc->watchdog.runlistId);
+            pKernelRc->watchdog.notifierToken->info32);
     }
     SLI_LOOP_END;
 }
